#!/usr/bin/python
# -*- coding: iso-8859-2 -*-

# 07.02.2007
# Mateusz Styrczula

""" Python OpenGL gadget """

import sys
import os
import threading
from OpenGL.GLUT import *
from OpenGL.GL import *
from OpenGL.GLU import *

from misc import *
from primitives import *
from materials import *
from camera import *
from world import *




class MouseFunction:
    """ ... """
    none, pan, zoom = range (0, 3)



class MainGL:
    """ main program class """

    def __init__ (self):
        """ default constructor - program starting point """
        self.prepare_variables ()
        self.init_GLUT ()
        self.assign_functions ()
        self.set_misc ()
        self.set_light_params ()
        self.create_scene ()
        glutMainLoop ()


    def display (self):
        """ GLUT calls this method every time window's content needs to be redisplayed """
        glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        glLoadIdentity ()
        self.camera.establish ()
        glTranslatef (self.tr_x, self.tr_y, self.tr_z)

        glLightfv (GL_LIGHT0, GL_POSITION, (20.0, 500.0, 20.0))
        self.scene.draw ()
        grid ()
        glFlush ()
        glutSwapBuffers ()


    def reshape (self, w, h):
        """ this method is called every time window is resized """
        glViewport (0, 0, w, h)
        self.camera.aspect = float (w) / float (h)
        self.camera.establish_projection ()


    def keyboard (self, key, x, y):
        """ this method is called every time any keyboard button is pressed """
        if key == 'q':
            self.scene.animation_thread.stop ()
            sys.exit ()
        elif key == '1':
            self.scene.animation_thread.restart ()
            glutIdleFunc (self.idle)
        elif key == '2':
            self.scene.animation_thread.pause ()
            glutIdleFunc (None)
        elif key == 'n':
            try:
                for o in self.scene.objects:
                    o.display_normals_switch ()
                glutPostRedisplay ()
            except:
                pass
        elif key == 'z':
            if self.polygon_fill:
                glPolygonMode (GL_FRONT_AND_BACK, GL_LINE)
            else:
                glPolygonMode (GL_FRONT_AND_BACK, GL_FILL)
            self.polygon_fill = not self.polygon_fill
            glutPostRedisplay ()
        elif key == 'x':
            if self.smooth:
                glShadeModel (GL_FLAT)
            else:
                glShadeModel (GL_SMOOTH)
            self.smooth = not self.smooth
            glutPostRedisplay ()
        elif key == 'w':
            self.tr_z += 5.0
            glutPostRedisplay ()
        elif key == 's':
            self.tr_z -= 5.0
            glutPostRedisplay ()
        elif key == 'a':
            self.tr_x += 5.0
            glutPostRedisplay ()
        elif key == 'd':
            self.tr_x -= 5.0
            glutPostRedisplay ()
        elif key == 'r':
            self.tr_y += 5.0
            glutPostRedisplay ()
        elif key == 'f':
            self.tr_y -= 5.0
            glutPostRedisplay ()


    def passive_motion (self, x, y):
        """ if all mouse buttons state are "UP" and movement is detected,
        this method's code is executed """
        self.last_mouse_x = x
        self.last_mouse_y = y


    def active_motion (self, x, y):
        """ when any mouse button is in state "DOWN" and movement is detected,
        this method's code is executed """
        if self.mouse_function == MouseFunction.pan:
            self.camera.theta.set_deg (float ((int (self.camera.theta.get_deg () + (float (y - self.last_mouse_y) / 1.0))) % 360))
            self.camera.phi.set_deg (float ((int (self.camera.phi.get_deg () + (float (x - self.last_mouse_x) / 1.0))) % 360))
        elif self.mouse_function == MouseFunction.zoom:
            self.camera.eye[2] += float (y - self.last_mouse_y)
            if self.camera.eye[2] < 10:
                self.camera.eye[2] = 10
            self.camera.establish_projection ()
        self.passive_motion (x, y)
        glutPostRedisplay ()


    def mouse (self, button, state, x, y):
        """ GLUT calls this method every time mouse button state is changed """
        self.passive_motion (x, y)
        if button == GLUT_LEFT_BUTTON:
            if state == GLUT_DOWN:
                self.mouse_function = MouseFunction.pan
            elif state == GLUT_UP:
                self.mouse_function = MouseFunction.none
        elif button == GLUT_RIGHT_BUTTON:
            if state == GLUT_DOWN:
                self.mouse_function = MouseFunction.zoom
            elif state == GLUT_UP:
                self.mouse_function = MouseFunction.none


    def idle (self):
        """ if animation thread is running - care about screen refreshes """
        self.scene.animation_thread.refresh_lock.acquire ()
        glutPostRedisplay ()
        self.scene.animation_thread.refresh_lock.release ()


    def prepare_variables (self):
        """ obiect fields initiation """
        self.last_mouse_x = 0
        self.last_mouse_y = 0
        self.tr_x = self.tr_y = self.tr_z = 0.0
        self.mouse_function = MouseFunction.none
        self.camera = Camera (fovy = Angle (45.0),
                              aspect = 4.0 / 3.0,
                              zNear = 10.0, zFar = 2000.0,
                              eye = [0.0, 0.0, 200.0],
                              center = [0.0, 0.0, 0.0],
                              up = [0.0, 1.0, 0.0])
        self.matlib = MaterialLibrary ()
        self.scene = GraphicScene ()
        self.scene.material_library = self.matlib
        self.polygon_fill = True
        self.smooth = True


    def init_GLUT (self):
        """ init GLUT, enable buffers, create window """
        glutInit (sys.argv)
        glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH)
        glutInitWindowSize (800, 600)
        glutInitWindowPosition ((1024/2-800/2), (768/2-600/2))
        glutCreateWindow ("(c) 2007 M.S.")
        #glutFullScreen ()


    def assign_functions (self):
        """ assign apropriate methods (callbacks) to GLUT calls """
        glutDisplayFunc (self.display)
        glutReshapeFunc (self.reshape)
        glutKeyboardFunc (self.keyboard)
        glutMotionFunc (self.active_motion)
        glutPassiveMotionFunc (self.passive_motion)
        glutMouseFunc (self.mouse)


    def set_misc (self):
        """ set "default" cursor, clear_color, shade_model,
            enable lighting and antialiasing """
        glutSetCursor (GLUT_CURSOR_LEFT_ARROW)
        glClearColor (0.2, 0.3, 0.5, 0.0)
        glShadeModel (GL_SMOOTH)
        glEnable (GL_LIGHTING)
        glLightModelf (GL_LIGHT_MODEL_LOCAL_VIEWER, 0.0)
        glLightModelf (GL_LIGHT_MODEL_TWO_SIDE, 0.0)
        glEnable (GL_DEPTH_TEST)
        glDepthFunc (GL_LEQUAL)
        glEnable (GL_CULL_FACE)
        glCullFace (GL_BACK)
        glFrontFace (GL_CCW)
        glEnable (GL_NORMALIZE)  # we're delivering unit normals, but when glScalef is called...
        glPolygonMode (GL_FRONT_AND_BACK, GL_FILL)
        glEnable (GL_BLEND)
        glEnable (GL_LINE_SMOOTH)
        glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glHint (GL_LINE_SMOOTH_HINT, GL_NICEST)
        glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST)
        

    def set_light_params (self):
        """ set global light parameters """
        # set "sunlight"
        glLightfv (GL_LIGHT0, GL_AMBIENT, (0.2, 0.2, 0.2, 1.0))
        glLightfv (GL_LIGHT0, GL_DIFFUSE, (1.0, 1.0, 1.0, 1.0))
        glLightfv (GL_LIGHT0, GL_SPECULAR, (1.0, 1.0, 1.0, 1.0))
        glLightfv (GL_LIGHT0, GL_POSITION, (20.0, 500.0, 20.0))
        glEnable (GL_LIGHT0)
        # set "viewlight"
        glLightfv (GL_LIGHT1, GL_AMBIENT, (0.1, 0.1, 0.1, 1.0))
        glLightfv (GL_LIGHT1, GL_DIFFUSE, (0.2, 0.2, 0.2, 1.0))
        glLightfv (GL_LIGHT1, GL_SPECULAR, (0.2, 0.2, 0.2, 1.0))
        glLightfv (GL_LIGHT1, GL_POSITION, (0.0, 0.0, 200))
        glEnable (GL_LIGHT1)



    def create_scene (self):
        """ intialize objects in scene """
        try:
            self.matlib.parse_file ("./meshes/scene.mats")
        except:
            pass
        try:
            self.scene.parse_geometry_file ("./meshes/scene.geom")
        except:
            pass
        try:
            self.scene.parse_animation_file ("./meshes/scene.anim")
        except:
            pass




# run the program!
if __name__ == "__main__":
    program = MainGL ()

