#!/usr/bin/python
# -*- coding: iso-8859-2 -*-

# 07.02.2007
# Mateusz Styrczula

""" base class graphic objects definitions """

import sys
import os
import math
import random
from string import *
from OpenGL.GLUT import *
from OpenGL.GL import *
from OpenGL.GLU import *

from misc import *
from angle import *




class DrawingWay:
    """ every graphic object can be wire or solid """

    wire, solid = range (2)




class Transform:
    """ store position, rotation and scale """

    def __init__ (self,
                  position = [0, 0, 0],
                  rotation_axis = [0, 0, 0],
                  rotation_angle = Angle (),
                  scale = [1, 1, 1]):
        """ default constructor """
        self.position = position
        self.rotation_axis = rotation_axis
        self.rotation_angle = rotation_angle
        self.scale = scale
        self.matrix = None


    def change_axes (self):
        """ MAX -> OpenGL axes conversion (call when needed) """
        self.position[1], self.position[2] = self.position[2], -(self.position[1])
        self.rotation_axis[1], self.rotation_axis[2] = self.rotation_axis[2], -(self.rotation_axis[1])
        self.scale[1], self.scale[2] = self.scale[2], self.scale[1] 


    def prepare_matrix (self):
        """ self.matrix = identity * position * rotation * scale """
        glPushMatrix ()
        glLoadIdentity ()
        glTranslatef (self.position[0], self.position[1], self.position[2])
        glRotatef (self.rotation_angle.get_deg (),
                   self.rotation_axis[0],
                   self.rotation_axis[1],
                   self.rotation_axis[2])
        glScalef (self.scale[0], self.scale[1], self.scale[2])
        self.matrix = glGetFloatv (GL_MODELVIEW_MATRIX)
        glPopMatrix ()




class Mesh:
    """ mesh readed from file """

    def __init__ (self, position = [0,0,0],
                        color = [0,0,0],
                        drawing_way = DrawingWay.solid,
                        material_library = None,
                        name = ""):
        """ default constructor """
        self.name = name
        self.parent = None
        self.children = []
        self.transform_frames = [Transform (position = position)]
        self.current_frame = 0
        self.material_name = None
        self.material_library = material_library
        self.norms_amount = self.verts_amount = self.faces_amount = 0
        self.normals = []
        self.verts = []
        self.faces_norms = []
        self.faces_verts = []
        self.display_normals = False
        # we'll be using two display lists
        self.display_mesh_list = glGenLists (2)
        self.display_normals_list = self.display_mesh_list + 1


    def display_normals_switch (self):
        """ do it recursively in all children """
        self.display_normals = not self.display_normals
        for c in self.children:
            c.display_normals_switch ()


    def change_axes (self):
        """ MAX -> OpenGL axes conversion """
        for normal in self.normals:
            normal[1], normal[2] = normal[2], -(normal[1])
        for vertex in self.verts:
            vertex[1], vertex[2] = vertex[2], -(vertex[1])


    def normalize_normals (self):
        """ normalize normals (make every normal one unit length) """
        self.normals = map (lambda v: normalize (v), self.normals)


    def get_size (self):
        """ returns size """
        return 0


    def inc_frame (self):
        """ increase current frame in self and all children """
        self.current_frame = (self.current_frame + 1) % len (self.transform_frames)
        for c in self.children:
            c.inc_frame ()


    def draw (self):
        """ draws object on screen """
        glPushMatrix ()
        glMultMatrixf (self.transform_frames[self.current_frame].matrix)
        glCallList (self.display_mesh_list)
        if self.display_normals:
            glCallList (self.display_normals_list)
        for c in self.children:
            c.draw ()
        glPopMatrix ()


    def prepare_call_lists (self):
        """ prepare call lists to draw object on screen """
        glNewList (self.display_mesh_list, GL_COMPILE)
        self.material_library.activate_material (self.material_name)
        glBegin (GL_TRIANGLES)
        # normals and triangles
        for face_nr in range (0, self.faces_amount):
            for i in range (0, 3):
                glNormal3f (self.normals[self.faces_norms[face_nr][i]][0],
                            self.normals[self.faces_norms[face_nr][i]][1],
                            self.normals[self.faces_norms[face_nr][i]][2])
                glVertex3f (self.verts[self.faces_verts[face_nr][i]][0],
                            self.verts[self.faces_verts[face_nr][i]][1],
                            self.verts[self.faces_verts[face_nr][i]][2])
        glEnd ()
        glEndList ()
        for c in self.children:
            c.prepare_call_lists ()
        # normals (for drawing)
        glNewList (self.display_normals_list, GL_COMPILE)
        glPushAttrib (GL_ENABLE_BIT)
        glDisable (GL_LINE_SMOOTH)
        glDisable (GL_BLEND)
        glMaterialfv (GL_FRONT, GL_AMBIENT, (0.0, 0.0, 0.0, 1.0))
        glMaterialfv (GL_FRONT, GL_DIFFUSE, (0.0, 0.0, 0.0, 1.0))
        glMaterialfv (GL_FRONT, GL_SPECULAR, (0.0, 0.0, 0.0, 1.0))
        for face_nr in range (0, self.faces_amount):
            for i in range (0, 3):
                glBegin (GL_LINES)
                glVertex3f (self.verts[self.faces_verts[face_nr][i]][0],
                            self.verts[self.faces_verts[face_nr][i]][1],
                            self.verts[self.faces_verts[face_nr][i]][2])
                v = map (lambda x,y: x+5*y,
                         self.verts[self.faces_verts[face_nr][i]],
                         self.normals[self.faces_norms[face_nr][i]])
                glVertex3f (v[0], v[1], v[2])
                glEnd ()
        glPopAttrib ()
        glEndList ()

