#!/usr/bin/python
# -*- coding: iso-8859-2 -*-

# 07.02.2007
# Mateusz Styrczula

""" 2D/3D scene containter """

import sys
import os
import math
import random
import threading
from OpenGL.GLUT import *
from OpenGL.GL import *
from OpenGL.GLU import *

from misc import *
from primitives import *




class AnimationThread (threading.Thread):
    """ animation thread """
    
    def __init__ (self, main):
        """ default constructor """
        threading.Thread.__init__ (self)
        self.sleep_time = 0.02
        self.thread_running = False
        self.thread_paused = True
        self.main = main
        self.lock = threading.Lock ()
        self.lock.acquire ()
        self.running_lock = threading.Lock ()
        self.refresh_lock = threading.Lock ()
        self.start ()
  

    def stop (self):
        """ stop thread and synchronize """
        if not self.thread_running:
            return None
        else:
            self.thread_running = False
            self.pause ()
            self.lock.release ()
            self.join ()


    def pause (self):
        """ pauses thread until call to restart """
        if not self.thread_paused:
            self.thread_paused = True
            self.running_lock.acquire ()
            self.running_lock.release ()


    def restart (self):
        """ safely restart this thread """
        if self.thread_paused:
            self.thread_paused = False
            self.lock.release ()
            self.lock.acquire ()


    def run (self):
        """ thread body """
        self.thread_running = True
        while self.thread_running:
            self.lock.acquire ()
            self.lock.release ()
            threading._sleep (self.sleep_time)
            self.running_lock.acquire ()
            while (not self.thread_paused):

                for o in self.main.objects:
                    o.inc_frame ()

                self.refresh_lock.acquire ()
                threading._sleep (self.sleep_time)
                self.refresh_lock.release ()
            self.running_lock.release ()
            self.thread_paused = True
        return None




class GraphicScene:
    """ class representing current scene """

    def __init__ (self):
        """ default constructor """
        self.objects = []
        self.material_library = None
        self.selected_object = ()
        self.so_last_coordinates = (0.0, 0.0, 0.0)
        self.minimum_distance = 0.0
        self.animation_thread = AnimationThread (self)


    def recursive_geometry_parse (self, in_file, parent):
        """ read one object (and - recursively - it's whole subtree) """
        try:
            # read mesh filename
            mesh_name = in_file.readline ()
            if mesh_name == "":
                return None       # None == EOF in this context
            # read mesh id
            mesh_id = int (in_file.readline ())
            # read mesh parent id (actually... it is not stored anywhere)
            mesh_parent_id = int (in_file.readline ())
            # read children uids (actually.. only children.count is needed...)
            line = in_file.readline ()
            tmp = split (line, ",")
            try:
                children_uids = map (lambda x: int (x), tmp)
            except:
                children_uids = []
            # read mesh position
            line = in_file.readline ()
            mesh_position = map (lambda x: float (x), split (line, ','))
            # create new mesh to write to
            mesh = Mesh (position = mesh_position,
                         material_library = self.material_library,
                         name = mesh_name[:-1])
            # set mesh parent
            mesh.parent = parent
            # add newly created mesh to parent's children list
            if parent != None:
                parent.children += [mesh]
            # read mesh rotation
            line = in_file.readline ()
            mesh_rotation = map (lambda x: float (x), split (line, ','))
            mesh.transform_frames[0].rotation_axis = mesh_rotation[0:3]
            mesh.transform_frames[0].rotation_angle.set_deg (mesh_rotation[3])
            # read mesh scale
            line = in_file.readline ()
            mesh.transform_frames[0].scale = map (lambda x: float (x), split (line, ','))
            # transform axes, and prepare transform matrix (in frame 0)
            mesh.transform_frames[0].change_axes ()
            mesh.transform_frames[0].prepare_matrix ()
            # read mesh material name
            mesh.material_name = (in_file.readline ())[:-1]
            # read amount of points and faces
            line = in_file.readline ()
            tmp = map (lambda x: int (x), split (line, ','))
            mesh.norms_amount = tmp[0]
            mesh.verts_amount = tmp[1]
            mesh.faces_amount = tmp[2]
            # read normals
            line = in_file.readline ()
            for s in map (lambda x: x[1:], split (line, '],')):
                if len (s) != 0:
                    mesh.normals += [map (lambda x: float (x), split (s, ','))]
            # read verts
            line = in_file.readline ()
            for s in map (lambda x: x[1:], split (line, '],')):
                if len (s) != 0:
                    mesh.verts += [map (lambda x: float (x), split (s, ','))]
            # read faces (normal indexes)
            line = in_file.readline ()
            for s in map (lambda x: x[1:], split (line, '],')):
                if len (s) != 0:
                    mesh.faces_norms += [map (lambda x: int (x) - 1, split (s, ','))]
            # read faces (vertex indexes)
            line = in_file.readline ()
            for s in map (lambda x: x[1:], split (line, '],')):
                if len (s) != 0:
                    mesh.faces_verts += [map (lambda x: int (x) - 1, split (s, ','))]
            # make MAX->GL axes change and normalization
            mesh.change_axes ()
            mesh.normalize_normals ()
            # parse all children
            for i in range (0, len (children_uids)):
                ret = self.recursive_geometry_parse (in_file, mesh)
                if ret == None:
                    return ret
        except Exception, arg:
            print ("recursive_geometry_parse: %s\n" % arg)
            raise Exception (arg)
        return mesh


    def parse_geometry_file (self, filename):
        """ fill points, faces, normals etc. from file (multi-objects) """
        try:
            # open file
            in_file = open (filename, 'r')
            # parse file
            while True:
                mesh = self.recursive_geometry_parse (in_file, None)
                if mesh == None:
                    break
                mesh.prepare_call_lists ()
                self.add_object (mesh)
            # close file
            in_file.close ()
        except Exception, arg:
            print ("%s: \"%s\"\n" % (arg, filename))
            sys.exit (-1)


    def recursive_animation_parse (self, in_file, mesh):
        """ read animation for one object (and - recursively - for it's whole subtree) """
        try:
            # read mesh filename
            mesh_name = (in_file.readline ())[:-1]
            if mesh_name == "":
                return None       # None == EOF in this context
            # read mesh id
            mesh_id = int (in_file.readline ())
            # primitive check if this is apropriate animation sequence
            if mesh_name != mesh.name:
                return None
            # reset animation sequence in mesh
            mesh.transform_frames = []
            mesh.current_frame = 0
            # read animation sequence
            line = (in_file.readline ())[:-2] # trunc last hash (#\n)
            frames = map (lambda frame: map (lambda transform: map (lambda x: float (x),
                                                                    split (transform, ',')),
                                             split (frame, ':')),
                          split (line, '#'))
            for frame in frames:
                transform = Transform (position = frame[0],
                                       rotation_axis = frame[1][0:3],
                                       rotation_angle = Angle (frame[1][3], Angle.deg),
                                       scale = frame[2])
                transform.change_axes ()
                transform.prepare_matrix ()
                mesh.transform_frames += [transform]
            # read animation for all children
            for c in mesh.children:
                ret = self.recursive_animation_parse (in_file, c)
                if ret == None:
                    return ret
        except Exception, arg:
            print ("recursive_animation_parse: %s\n" % arg)
            raise Exception (arg)
        return mesh


    def parse_animation_file (self, filename):
        """ for every object in scene fill it's transform_frames from file """
        try:
            # open file
            in_file = open (filename, 'r')
            # parse file
            for o in self.objects:
                mesh = self.recursive_animation_parse (in_file, o)
                if mesh == None:
                    break
            # close file
            in_file.close ()
        except Exception, arg:
            print ("%s: \"%s\"\n" % (arg, filename))
            sys.exit (-1)


    def add_object (self, o):
        """ add object to current scene """
        self.objects += [o]


    def count_objects (self):
        """ count objects in scene - it count only roots, actually """
        len (self.objects)


    def draw (self):
        """ draw all objects on screen """
        for o in self.objects:
            o.draw ()


    def is_object_selected (self):
        """ check if any object is selected """
        return (self.selected_object != ())


    def clear_object_selection (self):
        """ unselect selected object """
        if self.selected_object != ():
            self.selected_object = ()


    def select_object (self, coordinate):
        """ select object under given coordinate """
        pass


    def move_selected_object_to (self, x, y):
        """ this method is invoked when mouse_drag event occurs """
        if self.selected_object != ():
            self.so_last_coordinates = self.selected_object.position

